# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

NNCF_PACKAGE_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
HW_CONFIG_RELATIVE_DIR = "common/hardware/configs"

NNCF_CACHE_PATH = Path.home() / Path(".cache/nncf/")
CACHE_MODELS_PATH = NNCF_CACHE_PATH / Path("models")

# Environment variables below, if set, mark the execution environment
# so that certain actions within NNCF proper, such as telemetry event collection or
# debug dumps, are performed or not performed
NNCF_CI_ENV_VAR_NAME = "NNCF_CI"  # Must be set in CI environments
NNCF_DEV_ENV_VAR_NAME = "NNCF_DEV"  # Must be set in environments of the NNCF dev team machines
