# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.patterns.patterns import GraphPattern as GraphPattern
from nncf.common.graph.patterns.patterns import HWFusedPatternNames as HWFusedPatternNames
from nncf.common.graph.patterns.patterns import IgnoredPatternNames as IgnoredPatternNames
from nncf.common.graph.patterns.patterns import Patterns as Patterns
from nncf.common.graph.patterns.patterns import merge_two_types_of_operations as merge_two_types_of_operations
