# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from nncf.telemetry.extractors import CollectedEvent
from nncf.telemetry.extractors import TelemetryExtractor


class CompressionStartedWithQuantizeApi(TelemetryExtractor):
    def extract(self, _: Any) -> CollectedEvent:
        return CollectedEvent(name="compression_started", data="quantize_api")


class CompressionStartedWithQuantizeWithAccuracyControlApi(TelemetryExtractor):
    def extract(self, _: Any) -> CollectedEvent:
        return CollectedEvent(name="compression_started", data="quantize_with_accuracy_control_api")
